/* This file is generated by the layout maker tool. */
auto* const view__0 = createLogicalGroup(CRect(0, 0, 775, 515), -1, "", kCenterText, 14);
mainView = view__0;
auto* const view__1 = createBackground(CRect(0, 110, 775, 445), -1, "", kCenterText, 14);
imageContainer_ = view__1;
view__0->addView(view__1);
enterPalette(invertedPalette);
auto* const view__2 = createLogicalGroup(CRect(0, 0, 775, 110), -1, "", kCenterText, 14);
view__0->addView(view__2);
auto* const view__3 = createRoundedGroup(CRect(5, 4, 180, 105), -1, "", kCenterText, 14);
view__2->addView(view__3);
auto* const view__4 = createAboutButton(CRect(27, 5, 147, 65), kTagAbout, "", kCenterText, 14);
view__3->addView(view__4);
auto* const view__5 = createInfoButton(CRect(51, 69, 83, 101), kTagFirstChangePanel+kPanelInfo, "", kCenterText, 30);
panelButtons_[kPanelInfo] = view__5;
view__3->addView(view__5);
auto* const view__6 = createCCButton(CRect(92, 69, 124, 101), kTagFirstChangePanel+kPanelControls, "", kCenterText, 30);
panelButtons_[kPanelControls] = view__6;
view__3->addView(view__6);
auto* const view__7 = createSettingsButton(CRect(132, 69, 164, 101), kTagFirstChangePanel+kPanelSettings, "", kCenterText, 30);
panelButtons_[kPanelSettings] = view__7;
view__3->addView(view__7);
auto* const view__8 = createHomeButton(CRect(11, 69, 43, 101), kTagFirstChangePanel+kPanelGeneral, "", kCenterText, 30);
panelButtons_[kPanelGeneral] = view__8;
view__3->addView(view__8);
auto* const view__9 = createRoundedGroup(CRect(185, 5, 550, 105), -1, "", kCenterText, 14);
view__2->addView(view__9);
auto* const view__10 = createHLine(CRect(10, 35, 355, 40), -1, "", kCenterText, 14);
view__9->addView(view__10);
auto* const view__11 = createHLine(CRect(10, 70, 355, 75), -1, "", kCenterText, 14);
view__9->addView(view__11);
auto* const view__12 = createClickableLabel(CRect(10, 5, 280, 35), kTagLoadSfzFile, "DefaultInstrument.sfz", kLeftText, 20);
sfzFileLabel_ = view__12;
view__9->addView(view__12);
auto* const view__13 = createLabel(CRect(70, 40, 355, 70), -1, "", kLeftText, 20);
keyswitchLabel_ = view__13;
view__9->addView(view__13);
auto* const view__14 = createBadge(CRect(10, 42, 70, 68), -1, "", kCenterText, 20);
keyswitchBadge_ = view__14;
view__9->addView(view__14);
auto* const view__15 = createInactiveLabel(CRect(10, 40, 355, 70), -1, "No key switch", kLeftText, 20);
keyswitchInactiveLabel_ = view__15;
view__9->addView(view__15);
view__15->setVisible(false);
auto* const view__16 = createLabel(CRect(10, 73, 70, 98), -1, "Voices:", kRightText, 14);
view__9->addView(view__16);
auto* const view__17 = createPreviousFileButton(CRect(280, 10, 305, 35), kTagPreviousSfzFile, "", kCenterText, 24);
view__9->addView(view__17);
auto* const view__18 = createNextFileButton(CRect(305, 10, 330, 35), kTagNextSfzFile, "", kCenterText, 24);
view__9->addView(view__18);
auto* const view__19 = createChevronDropDown(CRect(330, 10, 355, 35), kTagFileOperations, "", kCenterText, 24);
fileOperationsMenu_ = view__19;
view__9->addView(view__19);
auto* const view__20 = createLabel(CRect(70, 73, 110, 98), -1, "", kCenterText, 14);
infoVoicesLabel_ = view__20;
view__9->addView(view__20);
auto* const view__21 = createLabel(CRect(130, 73, 170, 98), -1, "Max:", kRightText, 14);
view__9->addView(view__21);
auto* const view__22 = createLabel(CRect(170, 73, 205, 98), -1, "", kCenterText, 14);
numVoicesLabel_ = view__22;
view__9->addView(view__22);
auto* const view__23 = createLabel(CRect(245, 73, 305, 98), -1, "Memory:", kRightText, 14);
view__9->addView(view__23);
auto* const view__24 = createLabel(CRect(305, 73, 355, 98), -1, "", kCenterText, 14);
memoryLabel_ = view__24;
view__9->addView(view__24);
auto* const view__25 = createChevronValueDropDown(CRect(205, 77, 225, 97), kTagSetNumVoices, "", kCenterText, 16);
numVoicesSlider_ = view__25;
view__9->addView(view__25);
auto* const view__26 = createRoundedGroup(CRect(555, 5, 770, 105), -1, "", kCenterText, 14);
view__2->addView(view__26);
auto* const view__27 = createKnob48(CRect(32, 15, 80, 63), -1, "", kCenterText, 14);
view__26->addView(view__27);
view__27->setVisible(false);
auto* const view__28 = createValueLabel(CRect(27, 65, 87, 70), -1, "Center", kCenterText, 14);
view__26->addView(view__28);
view__28->setVisible(false);
auto* const view__29 = createKnobCCBox(CRect(5, 5, 75, 95), kTagSetCCVolume, "Volume", kCenterText, 14);
volumeCCKnob_ = view__29;
view__26->addView(view__29);
auto* const view__30 = createKnobCCBox(CRect(80, 5, 150, 95), kTagSetCCPan, "Pan", kCenterText, 14);
panCCKnob_ = view__30;
view__26->addView(view__30);
auto* const view__31 = createVMeter(CRect(155, 5, 178, 95), -1, "", kCenterText, 14);
meters_[0] = view__31;
view__26->addView(view__31);
auto* const view__32 = createVMeter(CRect(185, 5, 208, 95), -1, "", kCenterText, 14);
meters_[1] = view__32;
view__26->addView(view__32);
enterPalette(defaultPalette);
auto* const view__33 = createLogicalGroup(CRect(0, 110, 775, 445), -1, "", kCenterText, 14);
subPanels_[kPanelInfo] = view__33;
view__0->addView(view__33);
view__33->setVisible(false);
auto* const view__34 = createSquaredTransparentGroup(CRect(0, 0, 775, 335), -1, "", kCenterText, 14);
view__33->addView(view__34);
auto* const view__35 = createInfoLabel(CRect(20, 10, 80, 35), -1, "Curves:", kLeftText, 14);
view__34->addView(view__35);
auto* const view__36 = createInfoLabel(CRect(20, 35, 80, 60), -1, "Masters:", kLeftText, 14);
view__34->addView(view__36);
auto* const view__37 = createInfoLabel(CRect(20, 60, 80, 85), -1, "Groups:", kLeftText, 14);
view__34->addView(view__37);
auto* const view__38 = createInfoLabel(CRect(20, 85, 80, 110), -1, "Regions:", kLeftText, 14);
view__34->addView(view__38);
auto* const view__39 = createInfoLabel(CRect(20, 110, 80, 135), -1, "Samples:", kLeftText, 14);
view__34->addView(view__39);
auto* const view__40 = createInfoLabel(CRect(120, 10, 160, 35), -1, "0", kCenterText, 14);
infoCurvesLabel_ = view__40;
view__34->addView(view__40);
auto* const view__41 = createInfoLabel(CRect(120, 35, 160, 60), -1, "0", kCenterText, 14);
infoMastersLabel_ = view__41;
view__34->addView(view__41);
auto* const view__42 = createInfoLabel(CRect(120, 60, 160, 85), -1, "0", kCenterText, 14);
infoGroupsLabel_ = view__42;
view__34->addView(view__42);
auto* const view__43 = createInfoLabel(CRect(120, 85, 160, 110), -1, "0", kCenterText, 14);
infoRegionsLabel_ = view__43;
view__34->addView(view__43);
auto* const view__44 = createInfoLabel(CRect(120, 110, 160, 135), -1, "0", kCenterText, 14);
infoSamplesLabel_ = view__44;
view__34->addView(view__44);
auto* const view__45 = createLogicalGroup(CRect(0, 110, 775, 445), -1, "", kCenterText, 14);
subPanels_[kPanelControls] = view__45;
view__0->addView(view__45);
view__45->setVisible(false);
auto* const view__46 = createControlsPanel(CRect(0, 0, 775, 335), -1, "", kCenterText, 14);
controlsPanel_ = view__46;
view__45->addView(view__46);
auto* const view__47 = createLogicalGroup(CRect(0, 110, 775, 445), -1, "", kCenterText, 14);
subPanels_[kPanelSettings] = view__47;
view__0->addView(view__47);
auto* const view__48 = createSquaredGroup(CRect(0, 0, 775, 335), -1, "", kCenterText, 14);
view__47->addView(view__48);
auto* const view__49 = createValueMenu(CRect(180, 60, 250, 85), kTagSetOversampling, "", kCenterText, 14);
oversamplingSlider_ = view__49;
view__48->addView(view__49);
auto* const view__50 = createLabel(CRect(10, 60, 95, 85), -1, "Oversampling", kLeftText, 14);
view__48->addView(view__50);
auto* const view__51 = createLabel(CRect(10, 35, 85, 60), -1, "Preload size", kLeftText, 14);
view__48->addView(view__51);
auto* const view__52 = createValueMenu(CRect(180, 35, 250, 60), kTagSetPreloadSize, "", kCenterText, 14);
preloadSizeSlider_ = view__52;
view__48->addView(view__52);
auto* const view__53 = createValueMenu(CRect(430, 35, 510, 60), kTagSetSampleQuality, "", kCenterText, 14);
sampleQualitySlider_ = view__53;
view__48->addView(view__53);
auto* const view__54 = createLabel(CRect(270, 35, 320, 60), -1, "Sample", kLeftText, 14);
view__48->addView(view__54);
auto* const view__55 = createLabel(CRect(270, 85, 330, 110), -1, "Oscillator", kLeftText, 14);
view__48->addView(view__55);
auto* const view__56 = createValueMenu(CRect(430, 85, 510, 110), kTagSetOscillatorQuality, "", kCenterText, 14);
oscillatorQualitySlider_ = view__56;
view__48->addView(view__56);
auto* const view__57 = createOptionMenu(CRect(665, 35, 765, 60), kTagThemeMenu, "", kCenterText, 14);
themeMenu_ = view__57;
view__48->addView(view__57);
auto* const view__58 = createLabel(CRect(525, 35, 570, 60), -1, "Theme", kLeftText, 14);
view__48->addView(view__58);
auto* const view__59 = createLabel(CRect(525, 60, 620, 85), -1, "User SFZ folder", kLeftText, 14);
view__48->addView(view__59);
auto* const view__60 = createValueButton(CRect(665, 60, 765, 85), kTagChooseUserFilesDir, "DefaultPath", kCenterText, 14);
userFilesDirButton_ = view__60;
view__48->addView(view__60);
auto* const view__61 = createLabel(CRect(525, 200, 585, 225), -1, "Root key", kLeftText, 14);
view__48->addView(view__61);
auto* const view__62 = createLabel(CRect(525, 225, 595, 250), -1, "Frequency", kLeftText, 14);
view__48->addView(view__62);
auto* const view__63 = createStyledKnob(CRect(700, 250, 748, 298), kTagSetStretchedTuning, "", kCenterText, 14);
stretchedTuningSlider_ = view__63;
view__48->addView(view__63);
auto* const view__64 = createLabel(CRect(525, 260, 575, 285), -1, "Stretch", kLeftText, 14);
view__48->addView(view__64);
auto* const view__65 = createLabel(CRect(525, 175, 585, 200), -1, "Scala file", kLeftText, 14);
view__48->addView(view__65);
auto* const view__66 = createValueButton(CRect(624, 175, 724, 200), kTagLoadScalaFile, "DefaultScale", kCenterText, 14);
scalaFileButton_ = view__66;
view__48->addView(view__66);
auto* const view__67 = createValueMenu(CRect(685, 200, 720, 225), kTagSetScalaRootKey, "", kCenterText, 14);
scalaRootKeySlider_ = view__67;
view__48->addView(view__67);
auto* const view__68 = createValueMenu(CRect(720, 200, 750, 225), kTagSetScalaRootKey, "", kCenterText, 14);
scalaRootOctaveSlider_ = view__68;
view__48->addView(view__68);
auto* const view__69 = createResetSomethingButton(CRect(725, 175, 750, 200), kTagResetScalaFile, "", kCenterText, 14);
scalaResetButton_ = view__69;
view__48->addView(view__69);
auto* const view__70 = createLabel(CRect(270, 60, 415, 85), -1, "when freewheeling", kLeftText, 14);
view__48->addView(view__70);
auto* const view__71 = createValueMenu(CRect(430, 60, 510, 85), kTagSetFreewheelingSampleQuality, "", kCenterText, 14);
freewheelingSampleQualitySlider_ = view__71;
view__48->addView(view__71);
auto* const view__72 = createLabel(CRect(10, 85, 155, 110), -1, "Sustain cancels release", kLeftText, 14);
view__48->addView(view__72);
auto* const view__73 = createCheckbox(CRect(180, 85, 205, 110), kTagSetSustainCancelsRelease, "", kCenterText, 14);
sustainCancelsReleaseCheckbox_ = view__73;
view__48->addView(view__73);
auto* const view__74 = createLabel(CRect(270, 5, 380, 30), -1, "Rendering quality", kLeftText, 14);
view__48->addView(view__74);
auto* const view__75 = createHLine(CRect(10, 25, 250, 30), -1, "", kCenterText, 14);
view__48->addView(view__75);
auto* const view__76 = createLabel(CRect(10, 5, 105, 30), -1, "Engine options", kLeftText, 14);
view__48->addView(view__76);
auto* const view__77 = createLabel(CRect(525, 5, 565, 30), -1, "Other", kLeftText, 14);
view__48->addView(view__77);
auto* const view__78 = createLabel(CRect(525, 145, 750, 170), -1, "Tuning", kLeftText, 14);
view__48->addView(view__78);
auto* const view__79 = createHLine(CRect(270, 25, 509, 30), -1, "", kCenterText, 14);
view__48->addView(view__79);
auto* const view__80 = createHLine(CRect(525, 25, 765, 30), -1, "", kCenterText, 14);
view__48->addView(view__80);
auto* const view__81 = createHLine(CRect(525, 165, 750, 170), -1, "", kCenterText, 14);
view__48->addView(view__81);
auto* const view__82 = createChevronValueDropDown(CRect(725, 225, 750, 250), kTagSetTuningFrequency, "", kCenterText, 24);
tuningFrequencyDropdown_ = view__82;
view__48->addView(view__82);
auto* const view__83 = createTextEdit(CRect(665, 225, 725, 250), kTagSetTuningFrequency, "", kCenterText, 14);
tuningFrequencyEdit_ = view__83;
view__48->addView(view__83);
auto* const view__84 = createLabel(CRect(270, 110, 415, 135), -1, "when freewheeling", kLeftText, 14);
view__48->addView(view__84);
auto* const view__85 = createValueMenu(CRect(430, 110, 510, 135), kTagSetFreewheelingOscillatorQuality, "", kCenterText, 14);
freewheelingOscillatorQualitySlider_ = view__85;
view__48->addView(view__85);
auto* const view__86 = createLabel(CRect(525, 85, 620, 110), -1, "Zoom", kLeftText, 14);
view__48->addView(view__86);
auto* const view__87 = createOptionMenu(CRect(665, 85, 765, 110), kTagZoomMenu, "", kCenterText, 14);
zoomMenu_ = view__87;
view__48->addView(view__87);
auto* const view__88 = createValueButton(CRect(665, 110, 765, 135), kTagSetDefaultZoom, "Set Default", kCenterText, 14);
defaultZoomButton_ = view__88;
view__48->addView(view__88);
auto* const view__89 = createPiano(CRect(0, 445, 775, 515), -1, "", kCenterText, 16);
piano_ = view__89;
view__0->addView(view__89);
auto* const view__90 = createLogicalGroup(CRect(0, 110, 775, 445), -1, "", kCenterText, 14);
subPanels_[kPanelGeneral] = view__90;
view__0->addView(view__90);
view__90->setVisible(false);
enterPalette(invertedPalette);
auto* const view__91 = createHoverBox(CRect(5, 105, 175, 130), -1, "", kCenterText, 14);
lblHover_ = view__91;
view__0->addView(view__91);
view__91->setVisible(false);
enterPalette(defaultPalette);
