// SPDX-License-Identifier: BSD-2-Clause

// This code is part of the sfizz library and is licensed under a BSD 2-clause
// license. You should have receive a LICENSE.md file along with the code.
// If not, contact the sfizz maintainers at https://github.com/sfztools/sfizz

#pragma once

#include "utility/vstgui_before.h"
#include "vstgui/lib/vstguifwd.h"
#include "utility/vstgui_after.h"

using VSTGUI::CColor;

namespace gui {

    extern const CColor kColorTransparent;
    extern const CColor kColorTransparentDark;

    extern const CColor kColorOrange;

    extern const CColor kColorControlsScrollerTransparency;
    extern const CColor kColorControlsTransparency;
    extern const CColor kColorInfoTransparency;
    extern const CColor kColorMeterDanger;
    extern const CColor kColorMeterNormal;
    extern const CColor kColorTooltipBackground;
}
