/* This file is generated by the layout maker tool. */
auto* const view__0 = createaboutView(CRect(0, 0, 800, 475), -1, "About sfizz", kCenterText, 14);
aboutView = view__0;
auto* const view__1 = createLogo(CRect(200, 0, 600, 180), -1, "", kCenterText, 14);
view__0->addView(view__1);
auto* const view__2 = createInfoBox(CRect(260, 180, 540, 260), -1, "", kCenterText, 14);
view__0->addView(view__2);
auto* const view__3 = createDescriptionBox(CRect(200, 260, 600, 360), -1, "", kCenterText, 14);
view__0->addView(view__3);
auto* const view__4 = createHoverBox(CRect(460, 435, 775, 460), -1, "", kCenterText, 12);
lblHover_ = view__4;
view__0->addView(view__4);
view__4->setVisible(false);
auto* const view__5 = createButtonSfztools(CRect(460, 405, 500, 445), kTagButtonSfztools, "", kCenterText, 30);
view__0->addView(view__5);
auto* const view__6 = createButtonGithub(CRect(530, 405, 570, 445), kTagButtonGithub, "", kCenterText, 30);
view__0->addView(view__6);
auto* const view__7 = createButtonDiscord(CRect(600, 405, 640, 445), kTagButtonDiscord, "", kCenterText, 30);
view__0->addView(view__7);
auto* const view__8 = createButtonOpencollective(CRect(670, 405, 710, 445), kTagButtonOpencollective, "", kCenterText, 30);
view__0->addView(view__8);
auto* const view__9 = createButtonSfzformat(CRect(740, 405, 780, 445), kTagButtonSfzformat, "", kCenterText, 30);
view__0->addView(view__9);
auto* const view__10 = createSysInfoKeyBox(CRect(15, 375, 105, 465), -1, "Format\nHost\nProcessor\nAddress\nSystem", kLeftText, 14);
view__0->addView(view__10);
auto* const view__11 = createSysInfoValueBox(CRect(105, 375, 450, 465), -1, "%PluginFormat%\n%HostProgram%\n%HostCPU%\n%HostBits% bits\n%HostOS%", kLeftText, 14);
lblSysInfoValue_ = view__11;
view__0->addView(view__11);
